from chiplotle.geometry.core.coordinate import Coordinate
from chiplotle.geometry.core.path import Path
   
def get_line_intersection(line_a, line_b):
   '''
   Finds the intersection point, if any, between lines a and b.
   Returns a Coordinate or None if there is no intersection.
   '''
    
   #make sure we have two lines
   assert isinstance(line_a, Path) and len(line_a) == 2
   assert isinstance(line_b, Path) and len(line_b) == 2   
   
   p0_x, p0_y = line_a.points[0]
   p1_x, p1_y = line_a.points[1]
   p2_x, p2_y = line_b.points[0]
   p3_x, p3_y = line_b.points[1]
    
   s1_x = float(p1_x - p0_x)
   s1_y = float(p1_y - p0_y)
   s2_x = float(p3_x - p2_x)
   s2_y = float(p3_y - p2_y)

   s_divisor = (-s2_x * s1_y + s1_x * s2_y)
   if s_divisor == 0.0:
      s = -1.0
   else:
      s = (-s1_y * (p0_x - p2_x) + s1_x * (p0_y - p2_y)) / s_divisor
      
   t_divisor = (-s2_x * s1_y + s1_x * s2_y)
   
   if t_divisor == 0.0:
      t = -1.0
   else:
      t = ( s2_x * (p0_y - p2_y) - s2_y * (p0_x - p2_x)) / t_divisor

   if s >= 0 and s <= 1 and t >= 0 and t <= 1:
      #Collision detected
      i_x = p0_x + (t * s1_x)
      i_y = p0_y + (t * s1_y)
      return Coordinate(i_x, i_y)
   else:
      #No collision
      return None
   


## DEMO
if __name__ == '__main__':
   from chiplotle.geometry.shapes.line import line
   from chiplotle.geometry.core.group import Group
   from chiplotle.tools import io
   from random import randrange
   
   #draw a bunch of lines that do not intersect
   
   no_intersections = Group()

   line_1 = line([randrange(0, 4000), randrange(0, 4000)], 
      [randrange(0, 4000), randrange(0, 4000)])
   no_intersections.append(line_1)
   
   while len(no_intersections) < 300:
      new_line = line([randrange(0, 4000), randrange(0, 4000)], 
         [randrange(0, 4000), randrange(0, 4000)])
      
      intersection = False
      
      for l in no_intersections:
         if get_line_intersection(new_line, l) != None:
            intersection = True
            break
         
      if intersection == False:
         no_intersections.append(new_line)
         print "found %d lines..." % len(no_intersections)
               
   io.view(no_intersections)
   



from chiplotle.geometry.core.coordinate import Coordinate
from chiplotle.geometry.core.coordinatearray import CoordinateArray
from chiplotle.tools.geometrytools.get_minmax_coordinates import get_minmax_coordinates

def get_width_height(coords):
   '''Returns the width and height of the given `coords` list in a tuple.'''
   
   min_max = get_minmax_coordinates(coords.points)
   
   width = min_max[1][0] - min_max[0][0]
   height = min_max[1][1] - min_max[0][1]
   return (width, height)


## DEMO
if __name__ == '__main__':
   from chiplotle.geometry.shapes.circle import circle
   from chiplotle.geometry.shapes.random_walk_polar import random_walk_polar
   from chiplotle.geometry.core.group import Group
   from chiplotle.geometry.core.path import Path
   from chiplotle.tools import io

   #random points inside a circle
   rw = random_walk_polar(100)

   assert isinstance(rw, Path)
   
   (width, height) = get_width_height(rw)
   
   #a circle around all the points
   c = circle(width/2.0, 500)
   
   g = Group((rw, c))
   
   io.view(g)



from chiplotle.core.cfg.write_config_file import write_config_file
from chiplotle.core.cfg.write_log_file import write_log_file
from chiplotle.core.cfg.cfg import LOG_FILE
from chiplotle.core.cfg.cfg import CONFIG_FILE
from chiplotle.core.cfg.cfg import CONFIG_DIR
import os

def initialize_files():
   '''Initializes all needed directories and files.'''

   output = os.path.join(CONFIG_DIR, 'output')
   if not os.path.exists(output):
      raw_input('ATTENTION: "%s" does not exist in your system.\n\
      Chiplotle will now create it to store configuration and other\n\
      session files. \n\
      (Hit Return to continue)' % CONFIG_DIR)
      os.makedirs(output)

   if not os.path.exists(LOG_FILE):
      write_log_file(LOG_FILE)

   if not os.path.exists(CONFIG_FILE):
      raw_input('ATTENTION: Chiplotle has created its configuration file\n\
      %s.\n\
      Edit this file to set various Chiplotle defaults.' % CONFIG_FILE)
      write_config_file(CONFIG_FILE)




from chiplotle.core.interfaces.interface import _Interface

class ParentageInterface(_Interface):
   def __init__(self, _client):
      _Interface.__init__(self, _client)
      self.__parent = None

   ## PROPERTIES ##

   @property
   def parent(self):
      return self.__parent


   ## PRIVATE METHODS ##

   def _cut(self):
      '''Cut all ties with parent.'''
      self._cut_from_parent( )
      self._cut_parent( )

   def _cut_from_parent(self):
      '''Remove client from parent.'''
      client, parent = self._client, self.parent
      if parent is not None:
         parent._shapes.remove(client)

   def _cut_parent(self):
      '''Set parent of client to None.'''
      self.__parent = None

   def _switch(self, new_parent):
      '''Remove client from parent and give client to new parent.'''
      self._cut( )
      self.__parent = new_parent



from chiplotle.hpgl.commands import DI
from py.test import raises

def test_DI_01( ):
   t = DI( )

   assert t.run == t.rise == None
   assert t.format == 'DI;'


def test_DI_02( ):
   '''both run and rise (or none) must be set for formatting.'''

   t = DI(23)

   assert t.run == 23
   assert t.rise == None
   assert raises(Warning, 't.format')


def test_DI_03( ):
   '''run can be 0.'''

   t = DI(0, 25.3)

   assert t.run == 0
   assert t.rise == 25.3
   assert t.format == 'DI0.00,25.30;';




from chiplotle import *
from chiplotle.hpgl.commands import PA, EA

def test_get_bounding_box_01( ):
   '''The function can take a list of _HPGLPrimitives.'''
   t = hpgltools.get_bounding_box([PA([(1, 2), (3, 4)])])
   assert t == (Coordinate(1, 2), Coordinate(3, 4))


def test_get_bounding_box_02( ):
   '''The function can take a list of _HPGLPrimitives.'''
   a = [PA([(1, 2), (3, 4)]), EA((10, -2))]
   t = hpgltools.get_bounding_box(a)
   assert t == (Coordinate(1, -2), Coordinate(10, 4))






class _ChiplotleError(Exception):
   '''Abstract error class for Chiplotle!'''

   ## OVERRIDES ##

   def __eq__(self, arg):
      return self.__clas__.__name__ == arg.__class__.__name__

   def __repr__(self):
      return '%s( )' % self.__class__.__name__

   def __str__(self):
      return str(self.__class__.__name__)

## make Errors hashable?
#   def __hash__(self):
#      return hash(str(self))



class InitParameterError(_ChiplotleError):
   '''Error returned when initialization parameter is wrong 
   (e.g., wrong type, value, etc.').'''


class OperandError(_ChiplotleError):
   '''Error returned when an operand in a binary or unary operation
   is wrong (e.g., wrong type, value, etc.').'''




from chiplotle.geometry.shapes.circle import circle
from chiplotle.geometry.transforms.scale import scale

def symmetric_polygon_side_length(count, length):
   '''Creates a symmetric polygon with `count` sides, all with the 
   same given `length`.
   '''
   result = circle(1.0, count)
   scale(result, length / result.points.difference[0].magnitude)
   return result



from chiplotle.hpgl.commands import EW

def test_EW_01( ):

   t = EW(10, 0, 90)

   assert t.radius == 10
   assert t.startangle == 0
   assert t.sweepangle == 90
   assert t.chordangle is None
   assert t.format == 'EW10.00,0.00,90.00;'


def test_EW_02( ):

   t = EW(10, 0, 90, 4)

   assert t.radius == 10
   assert t.startangle == 0
   assert t.sweepangle == 90
   assert t.chordangle == 4
   assert t.format == 'EW10.00,0.00,90.00,4.00;'




from chiplotle.geometry.core.group import Group
from chiplotle.tools.mathtools.rotate_2d import rotate_coordinate_2d
from chiplotle.tools.geometrytools.scale import scale

class TransformLock(Group):

   def __init__(self, shapes, lock_transforms):
      Group.__init__(self, shapes = shapes)
      self.lock_transforms = set(lock_transforms)
      self.transform_map = {'scale' : self.scale_to_offset,
                            'rotate': self.rotate_to_offset,
                            }

   def get_transform(self, transform):
      fn = transform.func_name
      transform = self.transform_map.get(fn)
      if transform is None:
         raise ValueError("Don't know transform '%s'." % fn)
      return transform

   @staticmethod
   def rotate_to_offset(points, angle, pivot):
      def offset(points, value):
         return points + value
      center = points.center
      new_coord = rotate_coordinate_2d(center, angle, pivot)
      diff = new_coord - center
      return offset, [diff]

   @staticmethod
   def scale_to_offset(points, factor, pivot):
      def offset(points, value):
         return points + value
      #center = get_center(points)
      center = points.center
      new_coord = scale(center, factor, pivot)
      diff = new_coord - center
      return offset, [diff]


## ~~~~~~~~~~~~~~~~~~
if __name__ == '__main__':
   from chiplotle.geometry import shapes
   from chiplotle.geometry import transforms
   from chiplotle import io
   from chiplotle.hpgl import formatters
   import copy
   import math

   def rotation_example():
      r1 = shapes.rectangle(100, 100)
      r2 = shapes.rectangle(200, 200)
      transforms.offset(r2, (400, 0))
      r3 = shapes.rectangle(300, 300)
      transforms.offset(r3, (800, 0))

      tl = TransformLock([r1, r2], ['rotate'])
      start = Group([tl, r3])
      transforms.offset(start, (500, 0))

      mid = copy.deepcopy(start)
      end = copy.deepcopy(start)

      formatters.Pen(3)(start)
      formatters.Pen(2)(mid)
      formatters.Pen(1)(end)
      transforms.rotate(mid, math.pi / 6, (0, 0))
      transforms.rotate(end, math.pi / 3, (0, 0))

      o = shapes.circle(100)
      return Group([start, mid, end, o])


   def scale_example():
      s1 = shapes.star_outline(100, 100, 4)
      transforms.offset(s1, (300, 0))
      transforms.rotate(s1, -math.pi / 4)
      s2 = shapes.star_outline(100, 100, 6)
      transforms.offset(s2, (300, 0))
      transforms.rotate(s2, -math.pi / 2)
      s3 = shapes.star_outline(100, 100, 9)
      transforms.offset(s3, (300, 0))
      transforms.rotate(s3, -math.pi * 3 / 4)

      tl = TransformLock([s1, s2], ['scale'])
      start = Group([tl, s3])

      mid = copy.deepcopy(start)
      end = copy.deepcopy(start)

      formatters.Pen(3)(start)
      formatters.Pen(2)(mid)
      formatters.Pen(1)(end)
      transforms.scale(mid, 2)
      transforms.scale(end, 3)

      o = shapes.circle(100)
      return Group([start, mid, end, o])

   ## go...
   r = rotation_example()
   s = scale_example()

   io.view(Group([s, r]))



from chiplotle.tools.io.save_hpgl import save_hpgl
from chiplotle.core.cfg.cfg import CONFIG_DIR
import subprocess

def export(expr, filename, fmt = 'eps'):
   '''Export Chiplotle-HPGL objects to an image file format via ``hp2xx``.

   - `expr` can be an iterable (e.g., list) of Chiplotle-HPGL objects or a
      single Chiplotle-HPGL object.
   - `filename` the file name, including path but without extension.
   - `fmt` is a string describing the format of the file to which the 
      Chiplotle-HPGL objects will be exported. Default is 'eps'. 
      Valid formats are: jpg, png, tiff and many others. 
      Please see the ``hp2xx`` documentation for details.

   .. note::
      You must have ``hp2xx`` installed before you can export Chiplote-HPGL
      objects to image files.
   '''

   htmlfile = '{0}.hpgl'.format(filename)
   imgfile  = '{0}.{1}'.format(filename, fmt)
   save_hpgl(expr, htmlfile)

   cmd = 'hp2xx --truesize -p 1 -m %s -f "%s" "%s"' % (fmt, imgfile, htmlfile)
   p = subprocess.Popen(cmd, 
                        shell  = True,
                        stdout = subprocess.PIPE, 
                        stderr = subprocess.PIPE)
   stdout, stderr = p.communicate( )

   if 'not found' in stderr:
      print _hp2xxError()
   
   return imgfile


def _hp2xxError():
   msg = '''ATTENTION: hp2xx is not installed in your system.
         hp2xx must be installed for previewing art.'''
   return msg



from chiplotle.hpgl.abstract.arc import _Arc
from chiplotle.hpgl.abstract.hpglescape import _HPGLEscape
from chiplotle.hpgl.abstract.hpglprimitive import _HPGLPrimitive
from chiplotle.hpgl.abstract.penplot import _PenPlot
from chiplotle.hpgl.abstract.positional import _Positional
from chiplotle.hpgl.abstract.twopoint import _TwoPoint

class PU(_PenPlot):
   '''
   :Pen Up:
      Raises the pen from the plotting surface. Use this instruction 
      to prevent stray lines from being drawn.

   - `xy` : A ``list`` or ``tuple`` of x, y positions of the form \
   ``(x1, y2, x2, y2, x3, y3, ..., xn, yn)``.
   '''

   def __init__(self, xy=None):
      _PenPlot.__init__(self, xy)


class PD(_PenPlot):
   '''
   :Pen Down:
      Lowers the pen onto the writing surface for drawing and moves it 
      to the coordinates/increments you specified.

   - `xy` : A ``list`` or ``tuple`` of x, y positions of the form \
   ``(x1, y2, x2, y2, x3, y3, ..., xn, yn)``.
   '''

   def __init__(self, xy=None):
      _PenPlot.__init__(self, xy)


class PA(_PenPlot):
   '''
   :Plot Absolute:
      Establishes absolute plotting and moves the pen to specified 
      absolute coordinates using the current pen position.
   '''

   def __init__(self, xy=None):
      _PenPlot.__init__(self, xy)


class PR(_PenPlot):
   '''
   :Plot Relative:
      Establishes relative plotting and moves the pen (using the current 
      position) to the specified points, each successive move relative 
      to the last current pen location.
   '''

   def __init__(self, xy=None):
      _PenPlot.__init__(self, xy)


class CI(_HPGLPrimitive):
   '''
   :Circle:
      Draws a circle using the specified radius and chord tolerance. 
      If you want a filled circle, refer to the 
      :class:`~chiplotle.hpgl.commands.WG` or 
      :class:`~chiplotle.hpgl.commands.PM` instruction.
   '''
   _scalable = ['radius']

   def __init__(self, radius, chordangle=None):   
      self.radius = radius
      self.chordangle = chordangle

   @apply
   def radius( ):
      def fget(self):
         '''The radius of the circle.'''
         return self._radius
      def fset(self, arg):
         ### TODO: check for type here?
         self._radius = arg
      return property(**locals( ))

   @property
   def format(self):
      if self.chordangle:
         return '%s%.2f,%.2f%s' % (self._name, self.radius, self.chordangle, 
                              _HPGLPrimitive._terminator)
      else:
         return '%s%.2f%s' % (self._name, self.radius, _HPGLPrimitive._terminator)


class CC(_HPGLPrimitive):
   '''
   :Character chord angle:
      Sets the chord angle that determines the smoothness of characters
      drawn when you select one of the arc-font character sets for labeling.
   '''

   def __init__(self, angle=None):   
      self.angle = angle

   @property
   def format(self):
      if self.angle:
         return '%s%i%s' % (self._name, self.angle, _HPGLPrimitive._terminator)
      else:
         return '%s%s' % (self._name, _HPGLPrimitive._terminator)


class AF(_HPGLPrimitive):
   '''
   :Advance full page:
      Advances roll paper one full page length and establishes the 
      origin at the center of the new page.
   '''
   

class AH(_HPGLPrimitive):
   '''
   :Advance half page:
      Advances roll paper one half page length and establishes the 
      origin at the center of the new page.
   '''
   

class AP(_HPGLPrimitive):
   '''
   :Automatic Pen operations:
      Controls automatic pen operations such as returning a pen
      to the carousel if it has been in the holder without drawing
      for a certain time.

   For 7550:

   ====== ======= ===== =========================================
   bit_no dec_val state meaning
   ====== ======= ===== =========================================
   0        1      1    lift pen if down too long without motion
   0        0      0    do not lift pen until PU received
   1        2      1    put pen away if too long without  motion
   1        0      0    do not put pen away until SP0 received
   2        4      1    do not get new pen until drawing starts
   2        0      0    get pen immediately after SP command
   3        8      1    merge all pen up moves
   3        0      0    do not merge all pen up moves
   ====== ======= ===== =========================================

   default is 7 on 7550
   codes are 0 to 255 with default of 95 on the DraftMaster
   '''

   def __init__(self, n=None):   
      self.n = n

   @property
   def format(self):
      if self.n:
         return '%s%i%s' % (self._name, self.n, _HPGLPrimitive._terminator)
      else:
         return '%s%s' % (self._name, _HPGLPrimitive._terminator)
         

class AA(_Arc):
   '''
   :Arch Absolute:
      Draws an arc, using absolute coordinates, that starts at the
      current pen location and uses the specified center point.

   - `xy` : ``(x, y)`` position pair.
   - `angle` : ``float`` [-360 to 360]. The arch angle in degrees.
   - `chordtolerance` : ``float`` [0.36 to 180], ``None``.

   '''
   def __init__(self, xy, angle, chordtolerance=None):
      _Arc.__init__(self, xy, angle, chordtolerance)


class AR(_Arc):
   '''
   :Arch Relative:
      Draws an arc, using relative coordinates, that starts at the
      current pen location and uses the specified center point.

   - `xy` : ``(x, y)`` position pair.
   - `angle` : ``float`` [-360 to 360]. The arch angle in degrees.
   - `chordtolerance` : ``float`` [0.36 to 180], ``None``.
   '''

   def __init__(self, xy, angle, chordtolerance=None):
      _Arc.__init__(self, xy, angle, chordtolerance)


class AS(_HPGLPrimitive):
   '''
   :Acceleration Select:
      Sets pen acceleration for one or all pens. The default
      acceleration is suitable for all recommended pen and media
      combinations. Slowing the acceleration may improve line
      quality if you are using heavier than recommended media.

   - `accel` : ``int`` [1 to 4] , ``None``.  
   - `pen` : ``int`` [1 to 8], ``None``. When ``None``, accel is \
      applied to all pens.
   '''

   def __init__(self, accel=None, pen=None):   
      self.accel = accel
      self.pen = pen

   @property
   def format(self):
      if self.accel and self.pen:
         return '%s%i,%i%s' % (self._name, self.accel, self.pen, 
                              _HPGLPrimitive._terminator)
      elif self.accel:
         return '%s%i%s' % (self._name, self.accel, _HPGLPrimitive._terminator)  
      else:
         return '%s%s' % (self._name, _HPGLPrimitive._terminator) 


class EA(_Positional):
   '''
   :Edge Rectangle Absolute:
      Defines and outlines a rectangle using absolute coordinates.

   - `xy` : ``(x, y)``. The absolute coordinates of the remaining corner.
   '''

   def __init__(self, xy):
      _Positional.__init__(self, xy)


class ER(_Positional):
   '''
   :Edge Rectangle Relative:
      Defines and outlines a rectangle using relative coordinates.

   - `xy` : ``(x, y)``. The relative coordinates of the remaining corner.
   '''

   def __init__(self, xy):
      _Positional.__init__(self, xy)


class RA(_Positional):
   '''
   :Filled Rectangle Absolute:
      Defines and fills a rectangle using absolute coordinates.

   - `xy` : ``(x, y)`` tuple. The absolute coordinates of the \
      remaining corner.
   '''

   def __init__(self, xy):
      _Positional.__init__(self, xy)


class RR(_Positional):
   '''
   :Filled Rectangle Relative:
      Defines and fills a rectangle using relative coordinates.

   - `xy` : ``(x, y)`` tuple. The relative coordinates of the \
      remaining corner.
   '''

   def __init__(self, xy):
      _Positional.__init__(self, xy)


class VS(_HPGLPrimitive):
   ''' 
   :Pen Velocity:
      Set's pen velocity.

   - `vel` : ``float`` [0.0 to 127.9999] (depends on plotter), ``None``. 
   - `pen` : ``int`` [1 to 8].
   '''

   def __init__(self, vel=None, pen=None):
      self.vel = vel
      self.pen = pen
     
   @property
   def format(self):
      if self.vel and self.pen:
         return '%s%i,%i%s' % (self._name, self.vel, self.pen, 
            _HPGLPrimitive._terminator)
      elif self.vel:
         return '%s%i%s' % (self._name, self.vel, _HPGLPrimitive._terminator)
      else:
         return '%s%s' % (self._name, _HPGLPrimitive._terminator)


class FS(_HPGLPrimitive):
   '''
   :Force Select:
      Sets pen pressure to the paper for one or all pens. Use this 
      instruction to optimize pen life and line quality for each pen 
      and paper combination.

   - `force` : ``int`` [1 to 8]
   - `pen` : ``int`` [1 to 8]. If pen is ``None`` then all pens are set.
   '''

   def __init__(self, force=None, pen=None):
      self.force = force
      self.pen = pen

   @property
   def format(self):
      if self.force and self.pen:
         return '%s%i,%i%s' % (self._name, self.force, self.pen, 
                              _HPGLPrimitive._terminator)
      elif self.force:
         return '%s%i%s' % (self._name, self.force, _HPGLPrimitive._terminator)
      else:
         return '%s%s' % (self._name, _HPGLPrimitive._terminator)


class EP(_HPGLPrimitive):
   '''
   :Edge Polygon:
      Outlines the polygon currently stored in the polygon buffer. 
      Use this instruction to edge polygons that you defined in polygon mode 
      (:class:`~chiplotle.hpgl.commands.PM`) and with the rectangle and 
      wedge instructions (:class:`~chiplotle.hpgl.commands.RA`, 
      :class:`~chiplotle.hpgl.commands.RR` and 
      :class:`~chiplotle.hpgl.commands.WG`).
   '''


class BF(_HPGLPrimitive):
   '''
   Buffer plot.

   '''
   

class DC(_HPGLPrimitive):
   '''
   :Digitizer Clear:
      Terminates digitize mode. For example, if you are using an interrupt
      routine in a digitizing program to branch to another plotting function,
      use DC to clear the digitize mode immediately after branching. 
   '''
   

class DF(_HPGLPrimitive):
   '''
   :Default:
      Sets certain plotter functions to predefined default conditions.
      Use this instruction to return the plotter to a known state while 
      maintaining the current location of P1 and P2. When you use DF at 
      the beginning of a program, unwanted graphics parameters such as
      character size, slant, or scaling are not inherited from another
      program. 
   '''
   

class DP(_HPGLPrimitive):
   '''
   :Digitize Point:
      Returns the X,Y coordinates of a selected point on a plot to the
      computer for later use. Use this instruction to input data for a
      graphics program or to obtain the coordinates of a point or points
      on plot.
   '''
   

class FP(_HPGLPrimitive):
   '''
   :Fill Polygon:
      Fills the polygon currently in the polygon buffer. Use FP to fill
      polygons defined in polygon mode (
      :class:`~chiplotle.hpgl.commands.PM`) and defined with the edge 
      rectangle and wedge instructions (
      :class:`~chiplotle.hpgl.commands.EA`, 
      :class:`~chiplotle.hpgl.commands.ER`, and 
      :class:`~chiplotle.hpgl.commands.EW`).
   '''
   

class FR(_HPGLPrimitive):
   '''
   :Advance Frame:
      Advances paper to the next plot frame and calculates a relative 
      coordinate system for that frame. Use FR to do multi-frame long-axis 
      plotting.
   '''
   

class NR(_HPGLPrimitive):
   '''
   :Not Ready:
      Programmatically simulates pressing VIEW.
      However, you cannot take the plotter out of the view state with NR
      instruction.
   '''
   

class OA(_HPGLPrimitive):
   '''
   :Output Actual Pen Status:
      Outputs the current pen location (in plotter units) and up/down position.
      Use this information to position a label or figure, to determine the
      parameters of a window, or to determine the pen's curent location if you 
      moved it using front-panel cursor buttons.
   '''
   

class OC(_HPGLPrimitive):
   '''
   :Output Commanded Pen Status:
      Ouput the location and up/down position of the last commanded pen move 
      instruction. Use OC to position a label or determine the parameters of
      an instruction that tried to move the pen beyond the limits of some 
      window. You can also use this instruction when you want to know the 
      pen's location in user units.
   '''
   

class OD(_HPGLPrimitive):
   '''
   :Output Digitized Point and Pen Status:
      Outputs the X,Y coordinates and up/down pen position associated 
      with the last digitized point. Use this instruction after the 
      :class:`~chiplotle.hpgl.commands.DP` instruction to
      return the coordinates of the digitized point to your computer.
   '''
   

class OE(_HPGLPrimitive):
   '''
   :Output Error:
      Output a number corresponding to the type of HP-GL error (if any) 
      received by the plotter after the most recent 
      :class:`~chiplotle.hpgl.commands.IN` or 
      :class:`~chiplotle.hpgl.commands.OE` instruction. Use this 
      instruction for debugging programs. 
   
   =========  ========  ========================
   bit value  error no  meaning
   =========  ========  ========================
   0          0         no error
   1          1         unrecognized command
   2          2         wrong num of parameters
   4          3         out-of-range parameter
   8          4         unused
   16         5         unknown character set
   32         6         position overflow
   64         7         unused
   128        8         pinch wheels raised
   =========  ========  ========================
   
   .. note:: 
      some error meanings change depending on the plotter!
   '''
   

class OF(_HPGLPrimitive):
   '''
   :Output Factors:
      Outputs the number of plotter units per millimeter in each axis. 
      This lets you use the plotter with sofware that needs to know 
      the size of a plotter unit.
   '''
   

class OG(_HPGLPrimitive):
   '''
   :Output Group Count:
      Outputs the data block number of the current group count and 
      whether the escape function has been activated. Use this 
      instruction at the end of a data block in spooling applications, 
      where it is important to know the current data block number and 
      whether the data block has been transferred.
   '''


class OH(_HPGLPrimitive):
   '''
   :Output Hard-Clip Limits:
      Outputs the X,Y coordinates of the current hard-clip limits. 
      Use this instruction to determine the plotter unit dimension of 
      the area in which plotting can occur.
   '''
   

class OI(_HPGLPrimitive):
   '''
   :Output Identification:
      Outputs the plotter's identifying model number. This information is 
      useful in a remote operating configuration to determine which plotter 
      model is on-line, or when software needs the plotter's model number.
   ''' 


class OK(_HPGLPrimitive):
   '''
   :Output Key:
      Outputs a number that indicates which, if any, of the front-panel 
      function keys has been pressed. use this instruction with the 
      :class:`~chiplotle.hpgl.commands.WD` instruction when designing 
      interactive programs.
   '''
   

class OL(_HPGLPrimitive):
   '''
   :Output Label Length:
      Outputs information about the label contained in the label buffer.
   '''


class OO(_HPGLPrimitive):
   '''
   :Output Options:
      Outputs eight option parameters indicating the features implemented 
      on the plotter. Some software packages use this feature to determine 
      which plotter capabilities exist.
   '''
   

class OP(_HPGLPrimitive):
   '''
   :Output P1 and P2:
      Outputs the X,Y coordinates (in plotter units) of the current 
      scaling points P1 and P2. Use this instruction to determine the 
      numberic coordinates or P1 and P2 when they have been set manually, 
      and to help compute the number of plotter units per user units when 
      scaling is on.  This instruction can also be used with the input 
      window (:class:`~chiplotle.hpgl.commands.IW`) instruction to 
      programmatically set the window to P1 and P2.
   '''
   

class OS(_HPGLPrimitive):
   '''
   :Output Status:
      Outputs the decimal value of the status byte. Use this instruction in 
      debugging operations and in digitizing applications.

   =========  ============  =========
   bit value  bit position  meaning
   =========  ============  =========
   1          0             pen down
   2          1             P1 or P2 changed ("OP" clears)
   4          2             digitized point ready ("OD" clears)
   8          3             initialized ("OS" clears)
   16         4             ready to recieve data (always 0)
   32         5             There is an error ("OE" clears)
   64         6             unused
   128        7             unused
   =========  ============  =========
   
   power-on status == 24 (bits 3 & 4 set)
   '''
   

class OT(_HPGLPrimitive):
   '''
   :Output Carousel Type:
      Outputs information on the type of carousel loaded and the 
      stalls occupied.
   '''
   

class OW(_HPGLPrimitive):
   '''
   :Output Window:
      Outputs the X,Y coordinates of the lower-left and upper-right 
      corners of the window area in which plotting can occur. 
      This instruction is especially useful when the window area 
      (defined by :class:`~chiplotle.hpgl.commands.IW`) extends beyond 
      the hard-clip limits.
   '''
   

class PB(_HPGLPrimitive):
   '''
   :Print Buffer Label:
      Prints the contents of the label buffer.
   '''
   
      
class PS(_HPGLPrimitive):
   '''
   :Page Size:
      Changes the size of the hard clip limits.
   '''

   def __init__(self, length = None, width = None):
      self.length = length
      self.width = width

   @property
   def format(self):
      if self.length and self.width:
         return '%s%i,%i%s' % (self._name, self.length, self.width, 
            _HPGLPrimitive._terminator)
      elif self.length:
         return '%s%i%s' % (self._name, self.length, _HPGLPrimitive._terminator)
      else:
         return '%s%s' % (self._name, _HPGLPrimitive._terminator)


class BL(_HPGLPrimitive):
   '''
   :Buffer label:
      Stores a label in the label buffer. You can then use the
      output length (:class:`~chiplotle.hpgl.commands.OL`) instruction 
      to determine its space requirement prior to drawing it. Or, you 
      can use the plot buffer (:class:`~chiplotle.hpgl.commands.PB`)
      instruction to repeatedly plot this label.
   '''

   def __init__(self, label=None):
      self.label = label

   @property
   def format(self):
      if self.label:
         return '%s%s%s' % (self._name, chr(3), _HPGLPrimitive._terminator)
      else:
         return '%s%s' % (self._name, _HPGLPrimitive._terminator)


class IN(_HPGLPrimitive):
   '''
   :Initialize:
      Resets most plotter functions to their default settings. Use this 
      instruction to return the plotter to a known state and to cancel 
      settings that may have been changed by a previous program. 
   '''
   

class SS(_HPGLPrimitive):
   '''
   :Select standard character set:

   '''
   

class XT(_HPGLPrimitive):
   '''
   :X tick:

   '''
   

class YT(_HPGLPrimitive):
   '''
   :Y tick:
   
   '''
   

class CS(_HPGLPrimitive):
   '''
   :Standard character set:
      Designates a character set as the standard character set for labeling 
      instruction. Use this instruction to change the default ANSI ASCII 
      english set to one with characters appropriate to your application. 
      This instruction is particularly useful if you plot most of your
      labels in a language other than english.
   '''
   def __init__(self, set=0):   
      self.set = set

   @property
   def format(self):
      return '%s%i%s' % (self._name, self.set, _HPGLPrimitive._terminator)


class CT(_HPGLPrimitive):
   '''
   :Chord tolerance:
      Determines whether the chord tolerance parameter of the 
      :class:`~chiplotle.hpgl.commands.CI`, 
      :class:`~chiplotle.hpgl.commands.AA`, 
      :class:`~chiplotle.hpgl.commands.AR`
      and :class:`~chiplotle.hpgl.commands.WG` instructions is 
      interpreted as a chord angle in degrees or as a deviation distance 
      in current units.

   - `type` : ``int`` 0 or 1, default 0. 

   '''
   def __init__(self, type=0):   
      self.type = type

   @property
   def format(self):
      return '%s%i%s' % (self._name, self.type, _HPGLPrimitive._terminator)


class CV(_HPGLPrimitive):
   '''
   :Curved line generator:
      Collects coordinates (line segments) in the coordinate buffer so that they
      can be plotted as a group. This allows the plotter to plot in a
      continuous motion, rather than stopping and starting at each coordinate 
      endpoint. As a result, curves appear smoother. 
   
   - `n` : ``int`` 0 or 1, default 1 (on).
   - `inputdelay` : ``int`` [0 to 8,388,607] msec, default 100.
   '''

   def __init__(self, n=None, inputdelay=None):
      self.n = n
      self.inputdelay = inputdelay

   @property
   def format(self):
      if self.n and self.inputdelay:
         return '%s%i%i%s' % (self._name, self.n, self.inputdelay, 
         _HPGLPrimitive._terminator)
      elif self.n:
         return '%s%i%s' % (self._name, self.n, _HPGLPrimitive._terminator)
      else:
         return '%s%s' % (self._name, _HPGLPrimitive._terminator)
         

class CA(_HPGLPrimitive):
   '''
   :Designate alternate character set:
      Designates a character set as the alternate character set to be 
      used in labeling instructions. Use this instruction to provide an 
      additional character set that you can easily access in a program.
   
   - `set` : ``int`` [-1, 0 to 59, 60, 70, 80, 99, 100, 101], default 0.
   '''

   def __init__(self, set=0):   
      self.set = set

   @property
   def format(self):
      return '%s%i%s' % (self._name, self.set, _HPGLPrimitive._terminator)


class CM(_HPGLPrimitive):
   '''
   :Character selection mode:
      Specifies mode of character set selection and usage. Use this 
      instruction to select the alternate HP 8-bit, ISO 7-bit, or 
      ISO 8-bit character modes.

   - `switch` : ``int`` [0 to 3], default 0.
   - `fallback` : ``int`` 0 or 1, default 0.
   '''

   def __init__(self, switch=None, fallback=None):   
      self.switch = switch
      self.fallback = fallback

   @property
   def format(self):
      if self.switch and self.fallback:
         return '%s%i,%i%s' % (self._name, self.switch, self.fallback, 
         _HPGLPrimitive._terminator)
      elif self.switch:
         return '%s%i%s' % (self._name, self.switch, _HPGLPrimitive._terminator)
      else:
         return '%s%s' % (self._name, _HPGLPrimitive._terminator)


class CP(_HPGLPrimitive):
   '''
   :Character Plot:
      Move the pen the specified number of character plot cells from the
      current pen location.
   '''

   def __init__(self, spaces=None, lines=None):   
      self.spaces = spaces
      self.lines = lines

   @property
   def format(self):
      if self.spaces and self.lines:
         return '%s%s,%s%s' % (self._name, self.spaces, self.lines, 
         _HPGLPrimitive._terminator)
      elif self.spaces:
         return '%s%s%s' % (self._name, self.spaces, _HPGLPrimitive._terminator)
      else:
         return '%s%s' % (self._name, _HPGLPrimitive._terminator)


class DT(_HPGLPrimitive):
   '''
   :Define Label Terminator:
      Specifies the ASCII character to be used as the label terminator.
      Use this instruction to define a new label terminator if your 
      computer cannot use the default terminator (ETX, decimal code 3).
   '''

   def __init__(self, terminator=chr(3)):   
      self.labelterminator = terminator

   @property
   def format(self):
      return '%s%c%s' % (self._name, self.labelterminator, 
      _HPGLPrimitive._terminator)


class LB(_HPGLPrimitive):
   '''
   :Label:
      Plots text using the currently defined character set.
   '''

   def __init__(self, text):   
      self.text = text
      self.labelTerminator = chr(3)

   @property
   def format(self):
      return '%s%s%s%s' % (self._name, self.text, self.labelTerminator, 
         _HPGLPrimitive._terminator)


class SP(_HPGLPrimitive):
   '''
   :Select Pen:

   '''

   def __init__(self, pen = 0):   
      self.pen = pen

   @property
   def format(self):
      return '%s%i%s' % (self._name, self.pen, _HPGLPrimitive._terminator)


class LT(_HPGLPrimitive):
   '''
   :Line Type:
      Specifies the line pattern to be used when drawing linese and nonsolid
      fill types. Use LT to emphasize or de-emphasize other plotter lines and
      shapes.

   - `pattern` : ``int`` [-6 to 6]
   - `length` : ``float`` [0 to 100]

   =  ============================
   0  plot point at given point.
   1  .   .   .   .   .   .
   2  __   __   __   __   __
   3  ___ ___ ___ ___ ___
   4  __.__.__.__.__.__.
   5  ___ _ ___ _ ___ _ ___ _
   6  ___ _ _ ___ _ _ ___ _ _ ___
   =  ============================
   '''

   def __init__(self, pattern=None, length=4):   
      self.pattern = pattern
      self.length = length

   @property
   def format(self):
      if self.pattern is not None:
         return '%s%i,%.4f%s' % (self._name, self.pattern, 
         self.length, _HPGLPrimitive._terminator)
      else:
         return '%s%s' % (self._name, _HPGLPrimitive._terminator)


class FT(_HPGLPrimitive):
   '''
   :Fill Type:
      Selects the shading pattern used in polygons 
      (:class:`~chiplotle.hpgl.commands.FP`), rectangles 
      (:class:`~chiplotle.hpgl.commands.RA` or 
      :class:`~chiplotle.hpgl.commands.RR`), or wedges 
      (:class:`~chiplotle.hpgl.commands.WG`). Use this instruction to 
      enhance plots with solid fill, parallel lines (hatching), 
      cross-hatching, or a fill pattern you designed using the 
      user-defined fill type (UF) instruction.
   
   - `type` : ``int`` 1 or 2,  Solid (space and angle ignored) \
      3:  Hatching, 4:  Cross hatching.
   '''

   def __init__(self, type=None, space=None, angle=None):   
      self.type = type
      self.space = space
      self.angle = angle

   @property
   def format(self):
      if not None in (self.type, self.space, self.angle):
         return '%s%i,%s,%s%s' % (self._name, self.type, self.space,
         self.angle, _HPGLPrimitive._terminator)
      elif not None in (self.type, self.space):
         return '%s%i,%s%s' % (self._name, self.type, self.space, 
         _HPGLPrimitive._terminator)
      elif not self.type is None:
         return '%s%i%s' % (self._name, self.type, _HPGLPrimitive._terminator)
      elif None == self.type == self.space == self.angle:
         return '%s%s' % (self._name, _HPGLPrimitive._terminator)
      else:
         ### TODO: raise this type of warning in all other commands where
         ### this may be necessary.
         raise(Warning("Can't format %s with given parameters." % self._name)) 


class PM(_HPGLPrimitive):
   '''
   :Polygon Mode:
      Enter polygon mode for defining shapes such as block letters, 
      logos, surface charts, or any unique or intricate area for 
      subsequent filling and/or edging. Fill polygons using the fill 
      polygon (FP) instruction and/or outline them using the edge polygon 
      (EP) instruction.
   '''
   def __init__(self, n = 0):   
      self.n = n

   @property
   def format(self):
      return '%s%i%s' % (self._name, self.n, _HPGLPrimitive._terminator)


class EC(_HPGLPrimitive):
   '''
   :Enable Cut Line:
      Draws a dashed cut line between 'pages' on roll paper to indicate 
      where to cut the paper. Used with 
      :class:`~chiplotle.hpgl.commands.AF`, 
      :class:`~chiplotle.hpgl.commands.AH` and 
      :class:`~chiplotle.hpgl.commands.PG` instructions.
   '''

   def __init__(self, n = 0):   
      self.n = n

   @property
   def format(self):
      return '%s%i%s' % (self._name, self.n, _HPGLPrimitive._terminator)


class PG(_HPGLPrimitive):
   '''
   :Page Feed:
      Advances roll paper one page length and establishes the plotter-unit 
      origin at the center of the new page.
   '''

   def __init__(self, n = None):   
      self.n = n

   @property
   def format(self):
      if self.n:
         return '%s%i%s' % (self._name, self.n, _HPGLPrimitive._terminator)
      else:
         return '%s%s' % (self._name, _HPGLPrimitive._terminator)


class GC(_HPGLPrimitive):
   '''
   :Group Count:
      Allows you to assign an arbitrary number that will be output by the 
      :class:`~chiplotle.hpgl.commands.OG` instruction. Use 
      :class:`~chiplotle.hpgl.commands.GC` with the 
      :class:`~chiplotle.hpgl.commands.OG` instruction to monitor the 
      successful transfer of data blocks in spooling applications.
   '''

   def __init__(self, count=None):   
      self.count = count

   @property
   def format(self):
      if not self.count is None:
         return '%s%i%s' % (self._name, self.count, _HPGLPrimitive._terminator)
      else:
         return '%s%s' % (self._name, _HPGLPrimitive._terminator)


### TODO implement these:
#class GM(_HPGLPrimitive):

#class GP(_HPGLPrimitive):

#class IM(_HPGLPrimitive):

class SL(_HPGLPrimitive):
   ''' 
   :Character Slant: 
      Argument is tan of desired angle.
   '''

   def __init__(self, tan = 0):   
      self.tan = tan

   @property
   def format(self):
      return '%s%.4f%s' % (self._name, self.tan, _HPGLPrimitive._terminator)


class SA(_HPGLPrimitive):
   '''
   :Select alternate character set:
   
   '''
   

class RO(_HPGLPrimitive):
   ''' 
   :Rotate coordinate system:
   
   '''

   def __init__(self, angle = 0):   
      self.angle = angle

   @property
   def format(self):
      return '%s%i%s' % (self._name, self.angle, _HPGLPrimitive._terminator)


class RP(_HPGLPrimitive):
   ''' 
   :Replot:  
   
   '''

   def __init__(self, n = 1):   
      self.n = n

   @property
   def format(self):
      return '%s%i%s' % (self._name, self.n, _HPGLPrimitive._terminator)


class SM(_HPGLPrimitive):
   ''' 
   :Symbol Mode:  
      Plots the char at each plotted point. 
      char can be any printing ascii char, except ';'
      Calling without an argument cancels symbol mode.
   '''

   def __init__(self, char = None):
      self.char = char

   @property
   def format(self):
      if self.char:
         return '%s%c%s' % (self._name, self.char, _HPGLPrimitive._terminator)
      else:
         return '%s%s' % (self._name, _HPGLPrimitive._terminator)


class SC(_TwoPoint):
   '''
   :Scale:
      Establishes a user-unit coordinate system by mapping user-defined
      values onto the scaling points P1 and P2. Thus, you can plot in 
      units convenient to your application. In addition, you can use this
      instruction to establish automatic isotropic scaling or to relocate
      the origin and set a specific ratio of plotter units to user units.
      
   .. note:: DraftMaster also has a more complex version of 'SC' that \
      is not implemented yet.
   '''

   def __init__(self, coords=None):
     _TwoPoint.__init__(self, coords) 


class IP(_TwoPoint):
   '''
   :Input P1 and P2:
      Allows you to establish new or default locations for the scaling 
      points P1 and P2. P1 and P2 are used by the scale instruction 
      (:class:`~chiplotle.hpgl.commands.SC`) 
      to establish user-unit scaling. 
      The :class:`~chiplotle.hpgl.commands.IP` instruction is often used to 
      ensure that a plot is always the same size, regardless of how P1 
      and P2 might have been set from the front panel or the size of media 
      loaded in the plotter.
   '''

   def __init__(self, coords=None):   
      _TwoPoint.__init__(self, coords) 


class IV(_HPGLPrimitive):
   '''
   :Invoke Character Slot:
      Invokes a character set slot into either the right or left half of 
      the in-use code table. Primarily used with ISO modes of character 
      selection.
   '''

   def __init__(self, slot=None, left=None):
      self.slot = slot
      self.left = left

   @property
   def format(self):
      if not None in (self.slot, self.left):
         return '%s%i%i%s' % (self._name, self.slot, self.left,
            _HPGLPrimitive._terminator)
      elif not self.slot is None:
         return '%s%i%s' % (self._name, self.slot, _HPGLPrimitive._terminator)
      elif self.slot == self.left == None:
         return '%s%s' % (self._name, _HPGLPrimitive._terminator)
      else:
         raise(Warning("Can't format %s with given parameters." % self._name)) 


class IW(_TwoPoint):
   '''
   :Input Window:
      Defines a rectangular area, or window, that establishes soft-clip 
      limits.  Subsequent programmed pen motion will be restricted to this 
      area. Use this instruction when you want to be sure that your plot 
      falls within a specified area.
   '''

   def __init__(self, coords=None):
      _TwoPoint.__init__(self, coords) 


### TODO this is the exact same pattern as that of all other commands with
### two parameters. Refactor.
class KY(_HPGLPrimitive):
   '''
   :Define Key:
      Assigns a predefined function to one of the frontal panel function 
      keys.  Use this instruction with the WD instruction when designing 
      interactive programs.
   '''

   def __init__(self, key=None, function=None):
      self.key = key
      self.function = left

   @property
   def format(self):
      if not None in (self.key, self.function):
         return '%s%i%i%s' % (self._name, self.key, self.function,
            _HPGLPrimitive._terminator)
      elif not self.key is None:
         return '%s%i%s' % (self._name, self.key, _HPGLPrimitive._terminator)
      elif self.key == self.function == None:
         return '%s%s' % (self._name, _HPGLPrimitive._terminator)
      else:
         raise(Warning("Can't format %s with given parameters." % self._name)) 


class PT(_HPGLPrimitive):
   '''
   :Pen Thickness:
      Determines the spacing between the parallel lines in solid fill 
      patterns, according to the pen tip thickness.

   - `thickness` : ``float`` [0.1 to 5] mm, default is 0.3mm.
   '''

   def __init__(self, thickness = 0.3):
      self.thickness = thickness

   @property
   def format(self):
      return '%s%.2f%s' % (self._name, self.thickness, _HPGLPrimitive._terminator)
      

class SI(_HPGLPrimitive):
   '''
   :Absolute character size:
      Specifies the size of labeling characters in centimeters. 
      Use this instruction to establish character sizing that is not 
      dependent on the settings of P1 and P2.

   - `width` : ``float`` [-110 to 110] cm, excluding 0. 
   - `height` : ``float`` [-110 to 110] cm, excluding 0. 
   '''

   def __init__(self, width = None, height = None):   
      assert width != 0
      assert height != 0

      self.width = width
      self.height = height

   @property
   def format(self):
      if self.width and self.height:
         return '%s%.2f,%.2f%s' % (self._name, self.width, self.height, 
            _HPGLPrimitive._terminator)
      elif None == self.width == self.height:
         return '%s%s' % (self._name, _HPGLPrimitive._terminator)
      else:
         raise(Warning("Can't format %s without all parameters." % self._name))


class SR(SI):
   '''
   :Relative character size:
      Specifies the relative size of characters as a percentage of the
      distance between P1 and P2. Use this instruction to establish
      relative character sizes so that if the P1/P2 distance changes,
      the character sizes adjust to occupy the same relative ammount of
      space.

   - `width` : ``float`` [-100 to 100] percent, excluding 0. 
   - `height` : ``float`` [-100 to 100] percent, excluding 0. 
   '''
   
      
class DI(_HPGLPrimitive):
   '''
   :Absolute direction:
      Specifies the direction in which labels are drawn, independent of
      P1 and P2 settings. Use this instruction to change labeling 
      direction when you are labeling line charts, schematic drawings, 
      blueprints, and survey boudaries.

   - `run` : ``float``. cos(angle)
   - `rise` : ``float``. sin(angle)
   '''

   def __init__(self, run=None, rise=None):
      self.run = run
      self.rise = rise

   @property
   def format(self):
      if not None in (self.run, self.rise):
         return '%s%.2f,%.2f%s' % (self._name, self.run, self.rise, 
         _HPGLPrimitive._terminator)
      elif None == self.run == self.rise:
         return '%s%s' % (self._name, _HPGLPrimitive._terminator)
      else:
         raise(Warning("Can't format %s without all parameters." % self._name))
         

class DR(DI):
   '''
   :Relative Direction:
      Specifies the direction in which labels are drawn relative to 
      the scaling points P1 and P2. Label direction is adjusted when 
      P1 and P2 change so that labels maintain the same relationship 
      to the plotted data. Use :class:`~chiplotle.hpgl.commands.DI` 
      if you want label direction to be independent or P1 and P2.

   - `run` : ``float``. cos(angle)
   - `rise` : ``float``. sin(angle)
   '''


### TODO: figure out how this works and implement.
#class DL(_HPGLPrimitive):
#   '''
#   Define Downloadable Character
#   Allows you to design characters and store them in a buffer for 
#   repeated use by character set -1.
#   SYNTAX: DL character number (,pen control), X-coordinate, Y-coordinate
#   (,..., (,pen control(,...,))); or DL character number; or DL;
#   '''
#   def __init__(self, charnumber, pencontrol, xy):
#      self.charnumber = charnumber
#      self.pencontrol = pencontrol
#      self.xy = xy


class DS(_HPGLPrimitive):
   '''
   :Designate Character Set into Slot:
      Designates up to four character sets to be immediately available for 
      plotting. Used with ISO character sets and modes.
   '''

   def __init__(self, slot=None, set=None):
      self.slot = slot
      self.set = set

   @property
   def format(self):
      if self.slot and self.set:
         return '%s%i,%i%s' % (self._name, self.slot, self.set, 
         _HPGLPrimitive._terminator)
      else:
         return '%s%s' % (self._name, _HPGLPrimitive._terminator)


class DV(_HPGLPrimitive):
   '''
   :Direction Vertical:
      Specifies vertical mode as the direction for subsequent labels.
      Use this instruction to 'stack' horizontal characters in a column.
      A carriage return and a line feed lace the next 'column' to the 
      left of the previous one. 
   '''

   def __init__(self, vertical=0):
      self.vertical = bool(vertical)

   @property
   def format(self):
      return '%s%i%s' % (self._name, self.vertical, _HPGLPrimitive._terminator)
         

class ES(_HPGLPrimitive):
   '''
   :Extra space:
      Adjust space between characters and lines of labels without affecting
      character size.

   - `charspace` : ``float``, ``None``. Spacing between characters.
   - `linespace` : ``float``, ``None``. Spacing between lines.

   Character and line spacing values add (or substract) a fraction of the 
   standard spacing. 0 is the standard, positive values increase
   space and negative values reduce space. 1 doubles the standard space,
   0.5 adds half the standard space, and -1 substracts the standar space,
   causing the characters to draw on top of each other.
   '''

   def __init__(self, charspace = None, linespace = None):
      self.charspace = charspace
      self.linespace = linespace

   @property
   def format(self):
      if not None in (self.charspace, self.linespace):
         return '%s%.2f,%.2f%s' % (self._name, self.charspace, self.linespace, 
            _HPGLPrimitive._terminator)
      elif not self.charspace is None:
         return '%s%.2f%s' % (self._name, self.charspace, _HPGLPrimitive._terminator)
      else:
         return '%s%s' % (self._name, _HPGLPrimitive._terminator)
         

class LO(_HPGLPrimitive):
   '''
   :Label Origin:
      Positions labels relative to current pen location. Use LO to center, 
      left justify, or right justify label. The label can be drawn above or
      below the current pen location and can also be offset by an amount equal 
      to 1/2 the character's width and height.
      
   - `origin` : ``int`` [1-9] or [11-19].
   '''

   def __init__(self, origin = 1):
      self.origin = origin

   @property
   def format(self):
      return '%s%i%s' % (self._name, self.origin, _HPGLPrimitive._terminator)


class EW(_HPGLPrimitive):
   '''
   :Edge Wedge:
      Outlines any wedge. Use these instructions to produce sectors of 
      a pie chart.

   - `radius` : ``float``. 
   - `startangle` : ``float`` [0 - 360] degrees.
   - `sweepangle` : ``float`` [0 - 360] degrees.
   - `chordangle` : ``float`` [0.36 - 50] degrees.
   '''

   _scalable = ['radius']

   def __init__(self, radius, startangle, sweepangle, chordangle=None):
      self.radius = radius
      self.startangle = startangle
      self.sweepangle = sweepangle
      self.chordangle = chordangle

   @property
   def format(self):
      if self.chordangle:
         return '%s%.2f,%.2f,%.2f,%.2f%s' % (self._name, self.radius, 
         self.startangle, self.sweepangle, self.chordangle, 
         _HPGLPrimitive._terminator)
      else:
         return '%s%.2f,%.2f,%.2f%s' % (self._name, self.radius, 
         self.startangle, self.sweepangle, _HPGLPrimitive._terminator)


class WG(EW):
   '''
   :Filled wedge:
   
   '''


class TL(_HPGLPrimitive):
   '''
   Length of ticks drawn with the XT and YT instructions.

   - `tp` : percentage of (P2y - P1y) for XT or (P2x - P1x) for YT. \
      Denotes portion above X-axis or to the right of the Y-axis when
       difference is positive.
   - `tn` : same as `tp` except denotes portion below the X-axis and to \
      the left of the Y-axis. 0.5 is default for both.
   '''

   def __init__(self, tp = 0.5, tn = 0.5):
      self.tp = tp
      self.tn = tn

   @property
   def format(self):
      return '%s%.4f,%.4f%s' % (self._name, self.tp, self.tn, _HPGLPrimitive._terminator)
        

class WD(_HPGLPrimitive):
   '''
   :Write to display:
   
   '''

   def __init__(self, text):
      self.text = text

   @property
   def format(self):
      return '%s%s%s' % (self._name, self.text, chr(3)) # no terminator??



### ESCAPES ###

class B(_HPGLEscape):
   '''
   :Escape output buffer space:
   
   '''
   

class On(_HPGLEscape):
   '''
   :On:
      Places the plotter in a programmed on-state.
      Instructs the plotter to interpret data as HPGL and DCI instructions, 
      rather than plotting the data stream as literal text characters.
   '''

   @property
   def _name(self):
      return '('
   

class Off(_HPGLEscape):
   '''
   :Off: 
      Places the plotter in a programmed off-state.
   '''
   @property
   def _name(self):
      return ')'


class ExtendedError(_HPGLEscape):
   '''
   :ExtendedError:
      Get RS-232-C related error message.

   =========  ========
   error num  meaning
   =========  ========
   0          no i/o error
   10         output request received while still processing previous one
   11         invalid byte received after escape sequence ("ESC.")
   12         invalid byte received as part of a device control instruction
   13         parameter out of range
   14         too many parameters received
   15         framing, parity, or overrun error
   16         input buffer overflow
   =========  ========
   
   '''
   @property
   def _name(self):
      return 'E'


class K(_HPGLEscape):
   '''
   :Abort command: 
      Tells the plotter to discard commands in its buffer.
   '''

class SetHandshakeMode(_HPGLEscape):
   '''
   :Set Handshake Mode:
      Set one of three standard handshakes.

   0 (none)
   1 (Xon-Xoff)
   2 (ENQ-ACK)
   3 (hardwire)
   '''
   def __init__(self, mode=None):   
      self.mode = mode

   @apply
   def mode( ):
      def fget(self):
         return self._mode
      def fset(self, mode):
         if not mode in (None, 0,1,2,3):
            raise ValueError('mode must be in (0,1,2,3).')
         self._mode = mode
      return property(**locals( ))
         
   @property
   def _name(self):
      return 'P'

   @property
   def format(self):
      if self.mode is None:
         return '%s.%s' % (self._escape, self._name)
      else:
         return '%s.%s%i' % (self._escape, self._name, self.mode)



from chiplotle import *
from chiplotle.geometry.shapes.circle import circle
from chiplotle.tools.geometrytools.get_minmax_coordinates \
   import get_minmax_coordinates
import py.test

def test_geometrytools_get_minmax_coordinates_01( ):
   '''Argument cannot be an int.'''
   s = 3
   assert py.test.raises(Exception, 't = get_minmax_coordinates(s)')

def test_geometrytools_get_minmax_coordinates_02( ):
   '''Returns a duple of Coordinates.'''
   s = circle(100)
   t = get_minmax_coordinates(s.points)
   assert t == (Coordinate(-100, -100), Coordinate(100, 100))




from chiplotle.geometry.core.coordinate import Coordinate
from chiplotle.hpgl.commands import PA, PR

def pr_to_pa(arg, starting_position=None):
   '''Converts a given PR into PA starting at `starting_position`.
   
      - `arg` is a PR object.
      - `starting_position` is a coordinate pair. Default is (0, 0).
   '''
   if not isinstance(arg, PR):
      raise TypeError('`arg` must be of type PR');
   
   if len(arg.xy) == 0:
      return PA( )

      
   last_abs = starting_position or Coordinate(0, 0)
   last_abs = Coordinate(*last_abs)
   #abs_coords = [last_abs]
   abs_coords = [ ]
   for p in arg.xy:
      last_abs = last_abs + p
      abs_coords.append(last_abs)
   return PA(abs_coords)



import serial
import time

def what_plotter_in_port(port, wait_time=10):
   '''Check if there's a powered-on plotter in `port` port.
   
   - `port` : a string with port path or name.
   - `wait_time` : ``int`` maximum time in seconds to wait for 
      plotter response.

   Returns the ID of the plotter found or None.'''
   
   assert isinstance(wait_time, int)
   if not isinstance(port, basestring):
      raise TypeError('`port` must be a string.')

   from chiplotle.tools.serialtools import instantiate_serial_from_config_file
   ser = instantiate_serial_from_config_file(port)
   try:
      ser.flushInput()
      ser.flushOutput()
      ser.write('OI;')
   except serial.serialutil.SerialException:
      return None

   t = time.time( )
   while time.time( ) - t < wait_time:
      if ser.inWaiting( ) > 0:
         try:
            id = ser.readline(eol='\r').strip('\r') # <-- old pyserial
         except:
            id = ser.readline().strip('\r')

         ## if not just a repeater...
         if id != 'OI;': 
            return id



from chiplotle.geometry.core.coordinatearray import CoordinateArray
from chiplotle.geometry.core.group import Group
from chiplotle.geometry.core.path import Path
from chiplotle.geometry.core.coordinate import Coordinate

def test_shapes_group_points_01( ):
   '''An empty Group has an empty CoordinateArray.'''
   t = Group( )
   assert t.points == CoordinateArray([ ])


def test_shapes_group_points_02( ):
   '''A flat Group has a CoordinateArray.'''
   t = Group([Path([(1, 2), (3, 4)]), Path([(4, 2), (4, 2)])])
   assert t.points == CoordinateArray([(1, 2), (3, 4), 
      (4, 2), (4, 2)])


def test_shapes_group_points_03( ):
   '''A non-flat Group has a CoordinateArray.'''
   g = Group([Path([(1, 2), (3, 4)]), Path([(4, 2), (4, 2)])])
   t = Group([g, Path([(100, 200)])])
   assert t.points == CoordinateArray([(1, 2), (3, 4), (4, 2), (4, 2), (100, 200)])




from chiplotle.hpgl.commands import DR
from py.test import raises

def test_DR_01( ):
   t = DR( )

   assert t.run == t.rise == None
   assert t.format == 'DR;'


def test_DR_02( ):
   '''both run and rise (or none) must be set for formatting.'''

   t = DR(23)

   assert t.run == 23
   assert t.rise == None
   assert raises(Warning, 't.format')


def test_DR_03( ):
   '''run can be 0.'''

   t = DR(0, 25.3)

   assert t.run == 0
   assert t.rise == 25.3
   assert t.format == 'DR0.00,25.30;'


def test_DR_03( ):
   '''rise can be 0.'''

   t = DR(10, 0)

   assert t.run == 10
   assert t.rise == 0
   assert t.format == 'DR10.00,0.00;'




from chiplotle.tools.serialtools.scan_serial_ports import scan_serial_ports

## TODO: DELETE. No longer used.
def interactive_open_serial(baudrate, bytesize, parity, stopbits, timeout, 
   xonxoff, rtscts):
   '''The function scans for all serial ports, lists them and prompts 
   the user to select a port.  The function returns a Serial instance.
   '''
   ports = scan_serial_ports( )
   print "Available ports:\n"
   for k, port in ports.items( ):
      print "(%d) %s" % (k, port.portstr)
   ## get user's input...
   while True:
      sp = raw_input("\nChoose serial port number: ")
      ## check response...
      try:
         sp = int(sp)
         if not sp in ports.keys( ):
            raise ValueError
         else:
            break
      except ValueError:
         print 'Whoops! wrong port number. Try again...'
   ## configure port...
   selected_serial = ports[sp]
   selected_serial.baudrate = baudrate
   selected_serial.bytesize = bytesize
   selected_serial.parity = parity
   selected_serial.stopbits = stopbits
   selected_serial.timeout = timeout
   selected_serial.xonxoff = xonxoff
   selected_serial.rtscts = rtscts
   selected_serial.open( )
   print "Okay, opening %s." % selected_serial.portstr
   return selected_serial



from chiplotle.geometry.core.coordinate import Coordinate
from chiplotle.geometry.core.path import Path
import math

def arc_ellipse(width, height, 
                start_angle, end_angle, 
                segments = 100, segmentation_mode = '2PI'):  
   '''
   Constructs an arc from an ellipse with the given width, height,
   and number of segments. Arc goes from start_angle to end_angle,
   both of which are in radians.

      - `segmentation_mode` : '2PI' or 'arc'. The first segments
         the whole ellipse into the given number of segments,
         the second segments the arc.
   '''
   if start_angle > end_angle:
      end_angle += math.pi * 2
   
   def _divide_2pi():
      return (math.pi * 2) / float(segments)
   def _divide_arc():
      return (end_angle - start_angle) / float(segments)

   segmentation_map = {'2pi' : _divide_2pi,
                       'arc' : _divide_arc}

   rads_incr = segmentation_map[segmentation_mode.lower()]()
   rads = start_angle
   arc = []
   while rads < end_angle:
      coord = Coordinate(math.cos(rads), math.sin(rads))
      coord = coord * Coordinate(width / 2.0, height / 2.0)
      rads += rads_incr
      arc.append(coord)
   ## NOTE: this is better than using rads <= end_angle since
   ## the equality does not always work as expected with floats
   ## due to rounding.
   last = Coordinate(math.cos(end_angle), math.sin(end_angle))
   last *= Coordinate(width / 2.0, height / 2.0)
   arc.append(last)
   return Path(arc)


## RUN DEMO CODE

if __name__ == '__main__':
   from chiplotle.tools import io
   from chiplotle.geometry.core.group import Group
   
   gr = Group()
   for radius in range(100, 1000, 10):
       ae = arc_ellipse(radius, radius * 2, 0, math.pi/2, 5, 'arc')
       gr.append(ae)

   io.view(gr)




#!/usr/bin/env python

from chiplotle import *
from chiplotle.tools import *
import random
import sys

def abstract_masterpiece_file(file):

   saved = False

   while not saved:
      plot = generate_plot()

      io.view(plot)

      saveit = raw_input("\nsave this one (y/n)? ")
      if 'y' in saveit:
         io.save_hpgl(plot, file)
         print "saved " + file
         saved = True

   print "bye!"

def generate_plot():

   width = 30000
   height = 20000
   left = 0
   right = 30000
   bottom = 0
   top = 20000

   print "width: %d height: %d" % (width, height)
   pens = raw_input("\nhow many pens do you want to use? ")
   numPens = int(pens)
   
   plot = []
   
   #start in a random spot
   
   plot.append(PA([random.randint(left, right), random.randint(bottom, top)]))

   penNum = 1
   
   while True:
       #plotter.select_pen(penNum)
       plot.append(SP(penNum))

       whichGesture = random.randint(0,5)
       
       if whichGesture == 0:
           print "circle!"
           #plotter.circle(random.randint(10,5000), random.randint(1,180))
           x = random.randint(left, right)
           y = random.randint(bottom, top)
           plot.append(CI(random.randint(10,5000), random.randint(1,180)))
       
       
       elif whichGesture == 1:
           print "rect!"
           x = random.randint(left, right)
           y = random.randint(bottom, top)
           #plotter.edgeRectRelative(random.randint(10,5000), random.randint(10,5000))        
           plot.append(ER([random.randint(10,5000), random.randint(10,5000)]))

       elif whichGesture == 2:
           print "filled rect!"
           ft = random.randint(1,8)
           if ft == 1 or ft == 2:
               ft = 1
           if ft == 3 or ft == 4 or ft == 5:
               ft = 3
           if ft == 6 or ft == 7 or ft == 8:
               ft = 4
               
           space = random.randint(10,100)
           angle = random.randint(0,3) * 45
           
           print "fill type: %d space: %d angle: %d" % (ft, space, angle)
           #plotter.fill_type(ft, space, angle)
           #plotter.shadeRectRelative(random.randint(10,2000), random.randint(10,2000))
           plot.append(FT(ft, space, angle))
           plot.append(RR([random.randint(10,2000), random.randint(10,2000)]))
           

       elif whichGesture == 3:
           print "draw a crazy line!"
           #plotter.pen_down()
           #plotter.goto(random.randint(left, right), random.randint(bottom, top))
           #plotter.pen_up()
           plot.append(PD())
           plot.append(PA([random.randint(left, right), random.randint(bottom, top)]))
           plot.append(PU())
           
           
       elif whichGesture == 4:
           print "draw an abstract shape!"
           numPoints = random.randint(2,4)
           print "numPoints: ", numPoints
           firstX = random.randint(left, right)
           firstY = random.randint(bottom, top)
           #plotter.goto(firstX, firstY)
           #plotter.pen_down()
           plot.append(PA([firstX, firstY]))
           plot.append(PD())
           
           xRange = width/5
           yRange = height/5
                   
           for i in range(numPoints):
              #plotter.nudge(random.randint(int(-xRange), int(xRange)), random.randint(int(-yRange), int(yRange)))
              plot.append(PR([random.randint(int(-xRange), int(xRange)), random.randint(int(-yRange), int(yRange))]))
              
           
           #plotter.goto(firstX, firstY)
           #plotter.pen_up()
           
           plot.append(PA([firstX, firstY]))
           plot.append(PU())
           
       elif whichGesture == 5:
           print "just jump around!"
           #plotter.goto(random.randint(left, right), random.randint(bottom, top))
           plot.append(PA([random.randint(left, right), random.randint(bottom, top)]))
           
       #pick a new pen?
       pickPen = random.randint(0,99)
       if pickPen < 25:
           penNum += 1

       if penNum == numPens + 1:
           break
   
   plot.append(SP(0))
   
   return plot
   


### run main if called from command line like so: 
### $> python abstract_masterpiece.py
if __name__ == '__main__':
   if len(sys.argv) < 2:
      print 'Must give output filename.\nExample: $ abstract_masterpiece_file.py myfile.hpgl'
      sys.exit(2)
   file = sys.argv[1]

   abstract_masterpiece_file(file) 



#from chiplotle.core.imports.package_import import _package_import
#
#_package_import(__path__[0], locals( ))
#
#def remove_all_but_types(lst):
#   '''Keep only classes.'''
#   from types import TypeType
#   for k, v in lst.items( ):
#      if not isinstance(v, TypeType):
#         del(lst[k])
#
#remove_all_but_types(locals( ))

from dpx2000 import DPX2000
from dpx3300 import DPX3300
from dxy1300 import DXY1300
from dxy880 import DXY880
from hp7475a import HP7475A
from hp7550a import HP7550A
from hp7575a import HP7575A
from hp7576a import HP7576A
from hp7585b import HP7585B
from hp7595a import HP7595A
from hp7596a import HP7596A
from plotter import Plotter



## TODO: make this autoimport work
#from chiplotle.core.imports.package_import import _package_import
#_package_import(__path__[0], globals( ))

#from .get_centroid import get_centroid
#from .get_line_intersection import get_line_intersection



from chiplotle import *
from chiplotle.hpgl.commands import PU
from py.test import raises

### INITIALIZATION ###

def test_pu_init_01( ):
   '''PU can be initialized with nothing.'''
   t = PU( )
   assert type(t.xy) == CoordinateArray
   assert len(t.xy) == 0
   assert len(t.x) == 0
   assert len(t.y) == 0

def test_pu_init_02( ):
   '''PU can be initialized with empty list.'''
   t = PU([ ])
   assert type(t.xy) == CoordinateArray
   assert len(t.xy) == 0
   assert len(t.x) == 0
   assert len(t.y) == 0

def test_pu_init_03( ):
   '''PU argument must be list-like (list, tuple, Ndarray,...).'''
   assert raises(TypeError, 'PU(4)')

def test_pu_init_04( ):
   '''PU argument must be a list or tuple of length == 2*n'''
   assert raises(TypeError, 'PU([1])')
   assert raises(TypeError, 'PU([1,2,3])')
   assert raises(TypeError, 'PU([1,2,3,4,5])')

def test_pu_init_05( ):
   '''PU initialize properly with list or tuple of 2.'''
   t = PU([(1,2)])
   assert type(t.xy) == CoordinateArray
   assert t.xy == CoordinateArray([(1, 2)])
   assert t.x == (1, )
   assert t.y == (2, )

def test_pu_init_06( ):
   '''PU initialize properly with list or tuple of length == 2**n.'''
   t = PU([(1,2),(3,4)])
   assert type(t.xy) == CoordinateArray
   assert t.xy == CoordinateArray([(1, 2), (3, 4)])
   assert t.x == (1, 3)
   assert t.y == (2, 4)


### FORMAT ###

def test_pu_format_01( ):
   '''Empty PU( ).'''
   t = PU( )
   assert t.format == 'PU;'

def test_pu_format_02( ):
   '''Floats are left floats at format.'''
   t = PU([(1, 0.)])
   assert t.format == 'PU1.00,0.00;'

def test_pu_format_03( ):
   '''Ints are kept ints at format.'''
   t = PU([(0, 0)])
   assert t.format == 'PU0,0;'




from chiplotle import *
from chiplotle.hpgl.commands import ER
from py.test import raises

def test_ER_01( ):
   '''ER cannot initialize with a scalar.'''
   assert raises(TypeError, 't = ER(1)')


def test_ER_02( ):
   '''ER cannot initialize with a list of length > 2.'''
   assert raises(ValueError, 't = ER([1, 2, 3, 4])')


def test_ER_03( ):
   '''ER can initialize with a list of length 2.'''
   t = ER([1, 2])
   assert t.xy == Coordinate(1, 2)
   assert t.format == 'ER1,2;'



from chiplotle import *
from chiplotle.tools.plottertools import instantiate_plotters

plotter = instantiate_plotters( )[0]

assert plotter._buffer_space
assert plotter.id
assert plotter.actual_position
assert plotter.commanded_position
assert plotter.digitized_point
assert plotter.output_error
assert plotter.options
assert plotter.output_p1p2
assert plotter.status



from chiplotle.geometry.core.path import Path
import math
import random

def random_walk_cartesian(steps, step_size=500):
   '''Random Walk. Border is a square.'''

   the_points = []
   half_step = step_size/2
   
   previous = (0,0)
   
   for i in range(steps):
      x = previous[0] + random.randrange(-half_step, half_step)
      y = previous[1] + random.randrange(-half_step, half_step)      
      the_points.append((x, y))
   
   return Path(the_points)



## RUN DEMO CODE

if __name__ == '__main__':
   from chiplotle.tools import io

   rw = random_walk_cartesian(1000)
   assert isinstance(rw, Path)
   #print rw.format
   io.view(rw)


from chiplotle.geometry.core.path import Path
import math

def spiral_logarithmic(num_turns = 5, 
                       expansion_rate = 0.2, 
                       direction = "cw", 
                       segments = 500):  
   '''
   Constructs an logarithmic spiral with the given number of turns using the
   specified number of points.
   
   expansion_rate controls how large the spiral is for a given number of
   turns. Very small numbers will result in tightly-wound spirals. Large numbers
   will give spirals with giant "tails". Typical values range from 0.1 to 0.3
   
   The logarithmic spiral equation is:
   
   r = e^(bt)
   
   where r is the radius, e is e, b is the expansion rate, and t is theta
   
   '''
   
   two_pi = math.pi * 2.0
   total_rads = two_pi * num_turns
   theta = 0.0
   theta_incr = total_rads / float(segments - 1)

   expansion_rate = abs(expansion_rate)
   
   x, y = 0.0, 0.0
   spiral_points = [(x, y)]
   theta += theta_incr
   
   for i in range(segments - 1):
      x = math.cos(theta) * math.pow(math.e, (expansion_rate * theta))
      y = math.sin(theta) * math.pow(math.e, (expansion_rate * theta))

      if direction == "ccw":
         y *= -1.0
      
      spiral_points.append((x, y))
      theta += theta_incr
      
   return Path(spiral_points)


## RUN DEMO CODE

if __name__ == '__main__':
   from chiplotle.geometry.core.group import Group
   from chiplotle.geometry.shapes.line import line
   from chiplotle.geometry.transforms.offset import offset
   from chiplotle.tools import io
   
   s = spiral_logarithmic()

   #add some lines for scale
   line_right = line((0,0), (500, 0))
   line_left = line((0,0), (-500, 0))
   line_up = line((0,0), (0, 500))
   line_down = line((0,0), (0, -500))
   
   g = Group([s, line_right, line_left, line_up, line_down])

   io.view(g)




def difference(seq):
   '''Returns the difference between consecutive elements in `seq`.
   i.e., first derivative.
   '''
   result = [ ]
   for i in range(len(seq) - 1):
      result.append(seq[i+1] - seq[i])
   return type(seq)(result)



from chiplotle import *
import py.test


def test_scale_path_01():
   t = Path([(0, 0), (1, 1)])
   assert t.points == CoordinateArray([(0, 0), (1, 1)])
   scale(t, 2, Coordinate(0, 0))
   assert t.points == CoordinateArray([(0, 0), (2, 2)])

def test_scale_path_02():
   t = Path([(0, 0), (1, 1)])
   scale(t, 2, Coordinate(1, 1))
   assert t.points == CoordinateArray([(-1, -1), (1, 1)])




from chiplotle import *

## No longer the case
#def test_hpglescape_escape_01( ):
#   '''The escape class attribute of _HPGLEscape commands can be set directly
#   from the instances.'''
#   on = On( )
#   off = Off( )
#   on.escape = chr(1)
#   assert on.escape == chr(1)
#   assert off.escape == chr(1)
#   off.escape = chr(2)
#   assert on.escape == chr(2)
#   assert off.escape == chr(2)



from chiplotle import *
import copy

def test_perpendicular_noise_path_01( ):
   c  = circle(1000, 200)
   cc = copy.deepcopy(c)
   perpendicular_noise(c, 360)
   assert c.points != cc.points
   assert c.points is not cc.points




from chiplotle.hpgl import commands as hpgl
from chiplotle.tools.hpgltools.parse_hpgl_string import parse_hpgl_string
from chiplotle.tools.logtools.apply_logger import apply_logger
from chiplotle.tools.iterabletools.flat_list_to_pairs import flat_list_to_pairs

def inflate_hpgl_string(string, filter_commands=None):
   '''Reads a text string and "inflates" it by creating
   Chiplotle-HPGL class instances of the found HPGL commands.

   Example::

      chiplotle> sp = inflate_hpgl_string('SP1;')
      chiplotle> sp
      [SP(1)]

   Example::

      chiplotle> move = inflate_hpgl_string('IN;SP1;PA10,10;', ['IN'])
      chiplotle> move
      [SP(1), PA((10, 10))]
   '''

   filter_commands = filter_commands or [ ]

   if not isinstance(string, type('abc')):
      raise TypeError('`string` must be a string')
   if not isinstance(filter_commands, (list, tuple)):
      msg = '`filter_commands` must be a list of string HPGL commands.'
      raise TypeError(msg)

   _unsupported_commands = ('PW','PC', 'LA', 'WU', 'BP')
   comms = parse_hpgl_string(string)
   result = []
   for c in comms:
      if c: ## not an empty string: ''
         head = c[0:2]
         if head in filter_commands:
            continue
         if head in _unsupported_commands:
            continue
         command = inflate_hpgl_string_command(c)
         result.append(command)
   return result

@apply_logger
def inflate_hpgl_string_command(cmd_string):
   '''Converts a string representing a single HPGL command into a
   Chiplotle HPGL instance.
   e.g., 'PD1,2,3,4' --> PD((1,2,3,4)).
   '''
   head, body = _parse_hpgl_command_string(cmd_string)
   try:
      result = eval('hpgl.%s(%s)' % (head, body))
      return result
   except:
      msg = 'Could not create %s(%s)...' % (head, body)
      msg += ' The command is either malformed or unrecognized.'
      inflate_hpgl_string_command.logger.warning(msg)

def _parse_hpgl_command_string(cmd_string):
   '''Parses a single hpgl command string. 
   Splits it in two: the head and the body.
   e.g., 'PD1,2,3,4' --> 'PD' and '(1,2,3,4)'.
   '''
   head = cmd_string[0:2]
   if head in ('PU','PD','PA','PR','IP','IW','SC'):
      coords = cmd_string[2:].split(',')
      if coords == ['']:
         coords = []
      coords = [eval(n) for n in coords]
      coords = flat_list_to_pairs(coords)
      body = '(%s)' % coords
   elif head in ('RA','RR','ER','EA',):
      body = '(%s)' % cmd_string[2:]
   elif head in ('AR', 'AA'):
      parameters = cmd_string[2:].split(',')
      x = parameters.pop(0)
      y = parameters.pop(0)
      body = '(%s,%s),%s' % (x, y, ','.join(parameters))
   else:
      body = cmd_string[2:]
   return head, body


if __name__ == '__main__':
   string = 'IN;PU;PD1,2,3,4;'
   print inflate_hpgl_string(string)



from chiplotle.tools.hpgltools.inflate_hpgl_string import inflate_hpgl_string
from chiplotle.hpgl import commands as hpgl
import re

def import_hpgl_file(filename, filter_commands=None):
   '''Reads a text HPGL file and "inflates" it by creating
   Chiplotle-HPGL class instances of the found HPGL commands.

   Example::

      chiplotle> square = import_hpgl_file('examples/square.hpgl')
      chiplotle> square
      [SP(pen=1), PU(xy=[ 100.  100.]), PD(xy=[ 200.  100.]), 
      PD(xy=[ 200.  200.]), PD(xy=[ 100.  200.]), 
      PD(xy=[ 100.  100.]), SP(pen=0)]
   '''

   f = open(filename)
   fs = f.read()
   f.close()

   return inflate_hpgl_string(fs, filter_commands)



from chiplotle.geometry.core.label import Label

def label(text,
         charwidth, 
         charheight, 
         charspace = None, 
         linespace = None,
         origin = 'bottom-left'):  
   return Label(text, charwidth, charheight, charspace, linespace, origin)



from chiplotle.core.visitor import Visitor


class TagsVisitor(Visitor):
   '''Sort / group shapes by tags.'''
   
   def __init__(self):
      self.tags = {}

   def visit_Group(self, node, *args, **kwargs):
      self._sort_by_tag(node)
      for s in node:
         self.visit(s, *args, **kwargs)

   def visit__Shape(self, node, *args, **kwargs):
      self._sort_by_tag(node)
      

   ## private ##
   
   def _sort_by_tag(self, shape):
      for tag in shape.meta.tags: 
         if tag in self.tags:
            self.tags[tag].append(shape)
         else:
            self.tags[tag] = [shape]




from chiplotle import *
from chiplotle.hpgl.commands import *
from py.test import raises

def test_inflate_hpgl_string_01( ):
   '''An empty string returns an empty list.'''
   t = hpgltools.inflate_hpgl_string('')

   assert t == [ ]
  

def test_inflate_hpgl_string_02( ):
   '''The first argument must be a string.'''
   assert raises(TypeError, 't = hpgltools.inflate_hpgl_string([1,2,3])')


def test_inflate_hpgl_string_03( ):
   '''The function converts a string of HPGL commands into Chiplotle-HPGL instances.'''
   t = hpgltools.inflate_hpgl_string('IN;PU;PA10,10;CI100;')

   assert len(t) == 4
   assert t[0] == IN( )
   assert t[1] == PU( )
   assert t[2] == PA([(10, 10)])
   assert t[3] == CI(100)


def test_inflate_hpgl_string_04( ):
   '''HPGL commands need not be separated by ;.'''
   s = 'INPA10,10PDCI100'
   t = hpgltools.inflate_hpgl_string(s)
   assert len(t) == 4
   assert t[0] == IN( )
   assert t[1] == PA([(10, 10)])
   assert t[2] == PD( )
   assert t[3] == CI(100)


## AR AA ##

def test_inflate_hpgl_string_05( ):
   '''AA and AR are with only coordinates and angle are imported correctly.'''
   s = 'IN;AA10,10,90;'
   t = hpgltools.inflate_hpgl_string(s)
   assert len(t) == 2
   assert t[0] == IN( )
   assert t[1] == AA((10, 10), 90)


def test_inflate_hpgl_string_06( ):
   '''AA and AR are with all parameters are imported correctly.'''
   s = 'IN;AA10,10,90,10.2;'
   t = hpgltools.inflate_hpgl_string(s)
   assert len(t) == 2
   assert t[0] == IN( )
   assert t[1] == AA((10, 10), 90, 10.2)


## two point commands ##

def test_inflate_hpgl_string_IW():
   '''IW.'''
   s = 'IW1,2,3,4;'
   t = hpgltools.inflate_hpgl_string(s)
   assert len(t) == 1
   assert t[0] == IW([(1, 2), (3, 4)])

def test_inflate_hpgl_string_SC():
   '''SC.'''
   s = 'SC1,2,3,4;'
   t = hpgltools.inflate_hpgl_string(s)
   assert len(t) == 1
   assert t[0] == SC([(1, 2), (3, 4)])

def test_inflate_hpgl_string_IP():
   '''IP.'''
   s = 'IP1,2,3,4;'
   t = hpgltools.inflate_hpgl_string(s)
   assert len(t) == 1
   assert t[0] == IP([(1, 2), (3, 4)])

## FILTERING ##

def test_inflate_hpgl_string__filter_01( ):
   '''The function can take a list of HPGL commands to filter out from the result.'''
   t = hpgltools.inflate_hpgl_string('IN;PU;PA10,10;CI100;', ['IN', 'PA'])

   assert len(t) == 2
   assert t[0] == PU( )
   assert t[1] == CI(100)


def test_inflate_hpgl_string__filter_02( ):
   '''The `filter_commands' argument must be a list of two-letter HPGL strings.'''
   e = "t = hpgltools.inflate_hpgl_string('IN;PU;PA10,10;CI100;', 'IN;PU;')"

   assert raises(TypeError, e)
